#include <types.h>
#include <stdint.h>
#include <stddef.h>
#include <l4.h>
#include <gt.h>
#include <dev.h>
#include <l4e.h>

#include <assert.h>

/* For the u4600 IO map */
#include "u4600.h"
#include "u4600_io.h"
#include "gt64010a.h"

struct gt {
	mem_handle_t gt_addr;
	exec_handle_t int_handlers[NUM_PCI_INTERRUPTS];
	mem_handle_t cs1_handle;
	host_handle_t host;

	volatile char *ictrl;
};

/*
  Forward functipn declaration
*/

static void gt_conf_write_8(struct gt *gt, int bus, int device, 
			    int function, int register_num, uint8_t value);
static void gt_conf_write_16(struct gt *gt, int bus, int device, 
			    int function, int register_num, uint16_t value);
static void gt_conf_write_32(struct gt *gt, int bus, int device, 
			    int function, int register_num, uint32_t value);
static uint8_t gt_conf_read_8(struct gt *gt, int bus, int device, 
			       int function, int register_num);
static uint16_t gt_conf_read_16(struct gt *gt, int bus, int device, 
			       int function, int register_num);
static uint32_t gt_conf_read_32(struct gt *gt, int bus, int device, 
			       int function, int register_num);
static bus_space_t gt_pci_iospace_map(struct gt *gt, bus_addr_t address, 
				      size_t size, int flags);
static int  gt_pci_register_interrupt(struct gt *gt, int irq, 
				      exec_handle_t handler);

struct chipset_ops gt_ops = {
	(void*) gt_conf_write_8,
	(void*) gt_conf_write_16,
	(void*) gt_conf_write_32,
	(void*) gt_conf_read_8,
	(void*) gt_conf_read_16,
	(void*) gt_conf_read_32,
	(void*) gt_pci_iospace_map,
	(void*) gt_pci_register_interrupt
};

#ifdef DEBUG_GT
#define DEBUG(x...) l4e_printf(x)
#else
#define DEBUG(x...)
#endif

static void setup_pci_interrupts(struct gt *gt);
static int  gt_interrupt_event(struct gt *gt);

/******* GT STUFF *********/


/* Macros for accessing GT chip */
#define byte_swap32(x) \
	((unsigned int)( \
		(((unsigned int)(x) & (unsigned int)0x000000ffUL) << 24) | \
		(((unsigned int)(x) & (unsigned int)0x0000ff00UL) <<  8) | \
		(((unsigned int)(x) & (unsigned int)0x00ff0000UL) >>  8) | \
		(((unsigned int)(x) & (unsigned int)0xff000000UL) >> 24) ))

#define gtaddr32(a, x)	((volatile unsigned int   *)(a + (x)))
#define gtaddr16(a, x)	((volatile unsigned short *)(a + (x)))
#define gtaddr8(a, x)	((volatile unsigned char  *)(a + (x)))
#define gtval32(a, x)	byte_swap32(*gtaddr32(a, x))
#define gtval32set(a, x,y)	(*gtaddr32(a, x) = byte_swap32(y))
#define gtval32get(a, x)	(gtval32(a, x))

#define VALID_BUS(x) (x<16)

#define BUS_MASK  (0xFF)
#define BUS_SHIFT (16)
#define DEV_MASK  (0x1F)
#define DEV_SHIFT (11)
#define FUN_MASK  (0x07)
#define FUN_SHIFT (8)
#define REG_MASK  (0x3F)
#define REG_SHIFT (2)
#define CONFIG_ENABLE (1<<31)

#define VID_MASK  (0xFFFFU)
#define VID_SHIFT (0)
#define DID_MASK  (0xFFFF0000U)
#define DID_SHIFT (16)

static uint32_t
gt_read_config_reg( void *gt_addr, uint32_t bus, uint32_t device, 
		    uint32_t function, uint32_t reg )
{
	uint32_t addr, r;

	/* make up the Configuration address. pg 109 of the GT docs */
	addr = ((bus & BUS_MASK) << BUS_SHIFT)
		| ((device & DEV_MASK) << DEV_SHIFT)
		| ((function & FUN_MASK) << FUN_SHIFT)
		| ((reg & REG_MASK) << REG_SHIFT)
		| CONFIG_ENABLE;

	/* write into the address register */
	gtval32set( gt_addr, GT_PCI_CONF_ADDR, addr );

	/* read back the value from the device */
	r = gtval32get( gt_addr, GT_PCI_CONF_DATA );

	return r;
}

static void
gt_write_config_reg( void *gt_addr, uint32_t bus, uint32_t device, 
		     uint32_t function, uint32_t reg, uint32_t value )
{
	uint32_t addr;

	/* make up the Configuration address. pg 109 of the GT docs */
	addr = ((bus & BUS_MASK) << BUS_SHIFT)
		| ((device & DEV_MASK) << DEV_SHIFT)
		| ((function & FUN_MASK) << FUN_SHIFT)
		| ((reg & REG_MASK) << REG_SHIFT)
		| CONFIG_ENABLE;

	/* write into the address register */
	gtval32set( gt_addr, GT_PCI_CONF_ADDR, addr );

	/* write the value to the device */
	gtval32set( gt_addr, GT_PCI_CONF_DATA, value );
}

struct chipset *
gt_init(host_handle_t host)
{
	/* map us some memory for the GT chip */
	static struct chipset cs; /* FIXME: Should be a malloc - actually should be
				passed in... <sigh> */
	static struct gt gt; /* FIXME: Should be a malloc - actually should be
				passed in... <sigh> */
	gt.gt_addr = get_addr(map_physical(host, (void*)GT_BASE_ADDR, 0x1000));
	gt.host = host;
	/* Setup our PCI interrupt chip on the u4600 */
	setup_pci_interrupts(&gt);

	cs.ops = &gt_ops;
	cs.pvt = &gt;

	return &cs;
}


int
gt_register_interrupt_handler(struct gt *gt, int irq, void *handler, void *data)
{
	DEBUG("Registering handler for irq: %d for handler %p(%p)\n", 
	      irq, handler, data);

	/* We just use the host interrupts here */
	return register_cpu_interrupt_handler(gt->host, irq, handler, data);
}

/* Note: Assumes aligned */

uint32_t
gt_conf_read_32(struct gt *gt, int bus, int device, int function, 
		 int register_num )
{
	return gt_read_config_reg( gt->gt_addr, bus, device, 
				   function, (register_num >> 2) );
}

void
gt_conf_write_32(struct gt *gt, int bus, int device, int function, 
		  int register_num, uint32_t value )
{
	gt_write_config_reg( gt->gt_addr, bus, device, 
			     function, (register_num >> 2), value );
}

uint16_t gt_conf_read_16(struct gt *gt, int bus, int device, int function, 
			  int register_num )
{
	uint32_t r;

	r = gt_read_config_reg( gt->gt_addr, bus, device, 
				function, (register_num >>2) );

	if( (register_num >> 1) & 1 )
		r >>= 16;
	else
		r &= 0xffff;

	return (uint16_t) r;
}

void gt_conf_write_16(struct gt *gt, int bus, int device, int function, 
		       int register_num, uint16_t value )
{
	uint32_t r;

	/* read the register */
	r = gt_read_config_reg( gt->gt_addr, bus, device, 
				function, (register_num >>2) );

	if( (register_num >> 1) & 1 )
		r = (r & 0x0000ffff) | (((uint32_t)value) << 16);
	else
		r = (r & 0xffff0000) | value;

	/* write it back */
	gt_write_config_reg( gt->gt_addr, bus, device, 
			     function, (register_num >>2), r );
}

uint8_t 
gt_conf_read_8(struct gt *gt, int bus, int device, 
		int function, int register_num )
{
	uint32_t r;

	r = gt_read_config_reg( gt->gt_addr, bus, device, 
				function, (register_num >> 2) );

	if( register_num == 60 )
	{
		DEBUG( "gt: Someone is reading IRQ register!\n" );
		
		DEBUG( "gt: Writing a magic number over %x\n", r );
		gt_write_config_reg(gt->gt_addr, bus, device,
				    function, (register_num >> 2), r | 1);

		r = gt_read_config_reg(gt->gt_addr, bus, device, 
				       function, (register_num >> 2));
		DEBUG( "gt: With magic is %x\n", r );

		return device;
	}

	switch( register_num & 3 )
	{
		case 0:
			r &= 0xff;
			break;
		case 1:
			r >>= 8;
			r &= 0xff;
			break;
		case 2:
			r >>= 16;
			r &= 0xff;
			break;
		case 3:
			r >>= 24;
			break;
		default:
			DEBUG( "BER!\n" );
	}

	return (uint8_t) r;
}

static void 
gt_conf_write_8(struct gt *gt, int bus, int device, 
		 int function, int register_num, uint8_t value )
{
	uint32_t r;

	r = gt_read_config_reg(gt->gt_addr, bus, device, 
			       function, (register_num >> 2));

	if( register_num == 60 )
	{
		DEBUG( "gt: Someone is writing IRQ register! Ignoring!\n" );
		
		return;
	}

	switch( register_num & 3 )
	{
		case 0:
			r = (r & 0xffffff00) | value;
			break;
		case 1:
			r = (r & 0xffff00ff) | (((uint32_t)value)<<8);
			break;
		case 2:
			r = (r & 0xff00ffff) | (((uint32_t)value)<<16);
			break;
		case 3:
			r = (r & 0x00ffffff) | (((uint32_t)value)<<24);
			break;
		default:
			DEBUG( "BER!\n" );
	}

	/* write it back */
	gt_write_config_reg(gt->gt_addr, bus, device, 
			    function, (register_num >>2), r);
}

#if 0
/* IPCI_MemSpace interface */
bus_space_t 
cgt64111_ipci_memspace_space_map( cgt64111_t *cthis, bus_addr_t address, 
				  size_t size, int flags )
{
	bus_space_t rv;
	UserPrint("gt: memspace unimplemented, ignored\n");
	return rv;
}



/* IPCI_IOSpace interface */
bus_space_t 
cgt64111_ipci_iospace_space_map(struct gt *gt, bus_addr_t address, 
				size_t size, int flags)
{
	DEBUG( "Mapping address %p\n", address );
	address -= 0x10000000;

	return cgt64111_iisa_iospace_space_map( cthis, address, size, flags );
}
#endif

/* IPCI_Interrupts interface */
int 
gt_pci_register_interrupt(struct gt *gt, int irq, exec_handle_t handler)
{
	unsigned int real_int, reg;
	unsigned char c;

	/* maybe mask some stuff off... */
	assert(irq > 0);

	real_int = (irq-1)*4;  /* minus an extra one for u4600 wiring */

	DEBUG("gt: Registering interrupt handler %d\n", real_int);

	assert( real_int <= NUM_PCI_INTERRUPTS );

	gt->int_handlers[real_int] = handler;

	/* unmask the interrupt in the chip */
	if( real_int < 8 )
		reg = PCI_INT_MASK_LO;
	else
		reg = PCI_INT_MASK_HI;

	c = gt->ictrl[reg];
	DEBUG("gt: Masking int %d from %x\n", real_int, c);
	c &= ~(1 << (real_int%8));
	DEBUG("gt: ... to %x\n", c);

	/* write back the mask */
	gt->ictrl[reg] = c;

	return 0;
}

static bus_space_t
gt_pci_iospace_map(struct gt *gt, bus_addr_t address, 
		   size_t size, int flags )
{
	/* HACK HACK HACK */
	bus_space_t rv;

	rv = map_physical(gt->host, address, size);
	
	return rv;
}

#if 0
/* IISA_MemSpace interface */
bus_space_t 
cgt64111_iisa_memspace_space_map( cgt64111_t *cthis, bus_addr_t address, 
				  size_t size, int flags )
{
	/* HACK HACK HACK - this is a clag of above, which makes it WORSE! */
	/* Allocate PCI Mem0 */
	bus_space_t rv;
	static cap_t phys_cap = { NULL, 0 };
	int r;

	/* initially invalid */
	rv.address = phys_cap;
	rv.size = 32 * 1024 * 1024;  /* 32MB as per pg 29 of the GT docs */

	/* check if we have the IO space mapping - and make it */
	if( rv.address.address == NULL )
	{
		r = map_physical( (void*) 0x12000000, rv.size, cthis->node_cap,
				  &phys_cap);

		if( r != 0 )
		{
			UserPrint( "gt: Error getting Mem0 space mapping!\n" );
			return rv;
		}

		/* copy in the new value */
		rv.address = phys_cap;
	}

	/* add on the requested offset */
	rv.address.address += address;

	/* put in some cruft */
	rv.identifier = PYXIS_DENSE;
	rv.data = 0;
	rv.flags = 0;
	
	return rv;
}
#endif

#if 0
/* IISA_Interrupts interface */
int 
cgt64111_iisa_interrupts_register( cgt64111_t *cthis, 
				   cicap_t callback_instance, long interrupt )
{
	int rv;
	rv = (int)0;
	assert( !"isa_interrupt_map" );
	return rv;
}
#endif

#if 0
/* IRaw_Space interface */
bus_space_t
cgt64111_iraw_space_space_map(cgt64111_t *cthis, bus_addr_t address,
			      size_t size, int flags)
{
	static bus_space_t region;
	bus_space_t rv;
	int r;

	/* special hack for the GT timer component */
	if (address == GT_BASE_ADDR && size <= PAGESIZE) {
		rv.address = cthis->bus_cap;
		rv.size = size;
		rv.identifier = PYXIS_DENSE;
		rv.data = 0;
		rv.flags = 0;

		return rv;
	}

	/* this is mega-dodgy, and will break if mapping different regions
	 * FIXME... someday -AB */
	if (region.address.address == NULL) {
		r = map_physical((void *) CS0_BASE, BOOT_BASE - CS0_BASE,
				 cthis->node_cap, &region.address);
		assert(r == 0);

		/* put in some cruft */
		rv.size = BOOT_BASE - CS0_BASE;
		rv.identifier = PYXIS_DENSE;
		rv.data = 0;
		rv.flags = 0;
	}

	assert(address >= CS0_BASE && address + size < BOOT_BASE);
	rv = region;
	rv.size = size;
	rv.address.address = region.address.address + (address - CS0_BASE);

	return rv;
}
#endif

#if 0
/* IRaw_Interrupts interface */
int
cgt64111_iraw_interrupts_register(cgt64111_t *cthis,
				  cicap_t callback_instance, long interrupt)
{
	assert(interrupt != PCI_INTERRUPT_NUM);
	register_interrupt_handler(cthis, interrupt, callback_instance);
	return 0;
}
#endif


static void 
setup_pci_interrupts(struct gt *gt )
{
	int i;

	/* firstly clear the table of registered handlers */
	for( i = 0; i < NUM_PCI_INTERRUPTS; i++ ) {
		gt->int_handlers[i].function = NULL;
		gt->int_handlers[i].data = NULL;
	}
	/* give us access to the cs1 chip */
	DEBUG( "gt: Mapping CS1 space\n" );

	/* FIXME: This is u4600 only right now!! */
	/* see if the interrupt decoding chip is mapped */
	/* see pg 29 of the GT manual and u4600_io.h */
	gt->cs1_handle = map_physical(gt->host, 
				      (void*) CS1_BASE, 8 * 1024 * 1024);

	DEBUG( "gt: Poking at mask register...\n" );
	gt->ictrl = (char*) ((uintptr_t) get_addr(gt->cs1_handle) | PCI_INT_BITS);
	
	DEBUG( "gt: INT_LO  is %x\n", gt->ictrl[PCI_INT_INT_LO] );
	DEBUG( "gt: INT_HI  is %x\n", gt->ictrl[PCI_INT_INT_HI] );
	DEBUG( "gt: MASK_LO is %x\n", gt->ictrl[PCI_INT_MASK_LO] );
	DEBUG( "gt: MASK_HI is %x\n", gt->ictrl[PCI_INT_MASK_HI] );
	
	/* HACK HACK HACK: allow the IDE interrupts! */
	DEBUG( "gt: Allowing IDE...\n" );
	/* clear pending? :) */
	gt->ictrl[PCI_INT_INT_LO] = 0;
	gt->ictrl[PCI_INT_INT_HI] = 0;
	/* mask out everything */
	gt->ictrl[PCI_INT_MASK_LO] = 0xff;
	gt->ictrl[PCI_INT_MASK_HI] = 0xff;
	
	DEBUG("gt: INT_LO  is %x\n", gt->ictrl[PCI_INT_INT_LO]);
	DEBUG("gt: INT_HI  is %x\n", gt->ictrl[PCI_INT_INT_HI]);
	DEBUG("gt: MASK_LO is %x\n", gt->ictrl[PCI_INT_MASK_LO]);
	DEBUG("gt: MASK_HI is %x\n", gt->ictrl[PCI_INT_MASK_HI]);

	DEBUG("gt: PCI_INT_INT_LO  is %p\n", &gt->ictrl[PCI_INT_INT_LO]);
	DEBUG("gt: PCI_INT_INT_HI  is %p\n", &gt->ictrl[PCI_INT_INT_HI]);
	DEBUG("gt: PCI_INT_MASK_LO is %p\n", &gt->ictrl[PCI_INT_MASK_LO]);
	DEBUG("gt: PCI_INT_MASK_HI is %p\n", &gt->ictrl[PCI_INT_MASK_HI]);

	DEBUG("gt: Doing Mungi register\n");
	register_cpu_interrupt_handler(gt->host, PCI_INTERRUPT_NUM, 
				       gt_interrupt_event, gt);

}

#if 0
static void
call_handler( cgt64111_t *cthis, int handler )
{
	cap_t pb;
	IInterrupt_t *temp;
	environment_t env;
	void *obj;
	int r;
	cicap_t handler_cap;

	//UserPrint( "gt: Finding handler\n" );
	handler_cap = cthis->int_handlers[handler];

	if( handler_cap.ref == NULL )
	{
		UserPrint( "gt: bad registered handler: %d\n", handler );
		assert(!"No handler registered!");
	}

	//UserPrint( "gt: Creating obj\n" );
	obj = object_create(PAGESIZE, &pb);
	assert(obj != NULL);

	/* create the iinterrupt */
	//UserPrint( "gt: Doing cicap constructor\n" );
	temp = iinterrupt_cicap_constructor( handler_cap, pb,
					     PAGESIZE, &env);
	assert(temp != NULL);

	/* call the method */
	//UserPrint( "gt: Calling handler\n" );
	r = iinterrupt_interrupt_event( temp, &env );
	assert( r == 0 );

	//UserPrint( "gt: Handler called\n" );
}
#endif

static void
scan_ints(struct gt *gt, uint8_t bits, unsigned int offset )
{
	int i;

	/* quick check */
	if( bits == 0 )
		return;

	for( i = 0; i < 8; i++ )
	{
		/* if this interrupt is set */
		if ((bits & (1<<i)) != 0 ) {
			EXECUTE_HANDLE(gt->int_handlers[i]);
		}
	}
}

/* PCI Interrupt handler */
/* IInterrupt interface */
static int 
gt_interrupt_event(struct gt *gt )
{
	uint8_t lo, hi, masklo, maskhi;

	/* find out what it is */
	lo = gt->ictrl[PCI_INT_INT_LO];
	hi = gt->ictrl[PCI_INT_INT_HI];
	masklo = gt->ictrl[PCI_INT_MASK_LO];
	maskhi = gt->ictrl[PCI_INT_MASK_HI];

	/* mask them */
	lo &= ~masklo;
	hi &= ~maskhi;

	/* check for lo interrupts */
	scan_ints( gt, lo, 0 );

	/* check for hi interrupts */
	scan_ints( gt, hi, 8 );

	/* clear 'em all */
	/* FIXME: is there a reason not to clear them?
	 * should we mask off ones that aren't handled?
	 */
	gt->ictrl[PCI_INT_INT_LO] = 0;
	gt->ictrl[PCI_INT_INT_HI] = 0;
	
	return 0;
}

#if 0
/* IDMA interface implementation functions */
uint32_t
cgt64111_idma_get( cgt64111_t *cthis, cap_t cap, size_t size )
{
	uintptr_t addrp = (uintptr_t) cap.address;
	void *addr[4];
	uint32_t busaddr; 
	int i, r;

	/* XXX: Fixme, max pin of 32kb */
	assert(size <= 4 * PAGESIZE);

	for (i=0; i < (size / PAGESIZE); i++) {
		r = PagePin((void*) addrp, 1, &addr[i]);
		assert(r == 0);
		addrp += PAGESIZE;
	}

	busaddr = (uintptr_t)addr[0] + (uintptr_t)cap.address % PAGESIZE;

//	UserPrint("gt: idma: 0x%p 0x%lx -> 0x%p\n",
//		  cap.address, size, busaddr);

	return busaddr;
}

int
cgt64111_idma_release( cgt64111_t *cthis, uint32_t dma_addr )
{
	PageUnpin((void*)(uintptr_t)dma_addr, 4);
	return 0;
}
#endif
