/*
 * $Id: sosif.c,v 1.2 2003/09/10 11:36:22 benjl Exp $
 *
 * Copyright (c) 2001, Swedish Institute of Computer Science.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the Institute nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE INSTITUTE AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE INSTITUTE OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * This file is part of the lwIP TCP/IP stack.
 *
 * Author: Adam Dunkels <adam@sics.se>
 * Modified from stub for Mungi IDL version by <andrewb@cse.unsw.edu.au>
 */

#include <assert.h>
#include <string.h>
#include <stdlib.h>

#include <dev.h>

#include "lwip/debug.h"

#include "lwip/opt.h"
#include "lwip/def.h"
#include "lwip/mem.h"
#include "lwip/pbuf.h"
#include "lwip/sys.h"
#include "lwip/tcp.h"
#include "lwip/dhcp.h"
#include "netif/etharp.h"

#include "netif/sosif.h"

#include <tulip.h>

/* Define those to better describe your network interface. */
#define IFNAME0 'e'
#define IFNAME1 't'

static const struct eth_addr ethbroadcast = {{0xff,0xff,0xff,0xff,0xff,0xff}};

static void sosif_free_buffer(struct pbuf *p);

/*----------------------------------------------------------------------------*/
/*
 * low_level_output():
 *
 * Should do the actual transmission of the packet. The packet is
 * contained in the pbuf that is passed to the function. This pbuf
 * might be chained.
 *
 */
/*----------------------------------------------------------------------------*/
static err_t
low_level_output(struct netif *netif, struct pbuf *p)
{
	/* All this does is convert a pbuf into a 
	   packetsequence */

	pkt_seq packet_list;
	struct pbuf *q;
	packet_t packet;
	
	packet.c = 0;
	for (q=p; q != NULL; q= q->next) 
		if (q->len != 0) packet.c++;

	/* we are about to return, so send any pending packets */
	if (p == NULL) {
		return ERR_OK;
	}

	packet_list.c = 1;
	packet_list.v = &packet;

	/* Send the data from the pbuf to the interface, one pbuf at a time.
	 * The size of the data in each pbuf is kept in the ->len variable. */
	for (q = p; q != NULL; q = q->next) {
		
		if (q->len == 0)
			continue; /* catch zero-length pbufs */
		
		pbuf_ref(q);
		tulip_feedpacket(netif->state, (uint32_t) (uint64_t) q->payload, 
				 q->len, (q == p), (q->next != NULL), 
				 (uint32_t) (uint64_t) q);
	}

	return ERR_OK;
}

/* as above, but we need to work out the packet type manually */
static err_t
linkoutput(struct netif *netif, struct pbuf *p)
{
	struct eth_hdr *ethhdr;
	struct ip_hdr *iphdr;
	size_t offset;
	struct pbuf *q;

	assert(p->len >= sizeof(struct eth_hdr));
	ethhdr = p->payload;

#if 0
	UserPrint("link output dest %02x:%02x:%02x:%02x:%02x:%02x,",
		  ethhdr->dest.addr[0], ethhdr->dest.addr[1],
		  ethhdr->dest.addr[2], ethhdr->dest.addr[3],
		  ethhdr->dest.addr[4], ethhdr->dest.addr[5]);
	UserPrint(" src %02x:%02x:%02x:%02x:%02x:%02x,",
		  ethhdr->src.addr[0], ethhdr->src.addr[1],
		  ethhdr->src.addr[2], ethhdr->src.addr[3],
		  ethhdr->src.addr[4], ethhdr->src.addr[5]);
	UserPrint(" type 0x%hx\n", htons(ethhdr->type));
#endif

#if 0
	if (htons(ethhdr->type) != ETHTYPE_IP)
		return low_level_output(netif, p, PACKET_NONE);
#endif

	/* locate IP header within packet */
	offset = sizeof(struct eth_hdr);
	for (q = p; q != NULL; q = q->next)
		if (offset < q->len)
			break;
		else
			offset -= q->len;

	assert(q != NULL);
	assert(q->len - offset >= sizeof(struct ip_hdr));
	iphdr = q->payload + offset;

#if 0
	UserPrint("ver %hhd hl %hhd tod %hhd", IPH_V(iphdr), IPH_HL(iphdr),
		  IPH_TOS(iphdr));
	UserPrint(" totlen %hd id %hd off %hd", IPH_LEN(iphdr), IPH_ID(iphdr),
		  IPH_OFFSET(iphdr));
	UserPrint(" ttl %hhd proto %hhd checksum %hd\n", IPH_TTL(iphdr),
		  IPH_PROTO(iphdr), IPH_CHKSUM(iphdr));
	UserPrint("src %hhd.%hhd.%hhd.%hhd, dst %hhd.%hhd.%hhd.%hhd\n",
		  (u8_t)((iphdr->src.addr) >> 24) & 0xff,
		  (u8_t)((iphdr->src.addr) >> 16) & 0xff,
		  (u8_t)((iphdr->src.addr) >> 8) & 0xff,
		  (u8_t)(iphdr->src.addr & 0xff),
		  (u8_t)((iphdr->dest.addr) >> 24) & 0xff,
		  (u8_t)((iphdr->dest.addr) >> 16) & 0xff,
		  (u8_t)((iphdr->dest.addr) >> 8) & 0xff,
		  (u8_t)(iphdr->dest.addr & 0xff));
#endif

	//	return low_level_output(netif, p, get_packet_type(q->payload + offset));
	return 0;
}

/*----------------------------------------------------------------------------*/
/*
 * sosif_output():
 *
 * This function is called by the TCP/IP stack when an IP packet
 * should be sent. It calls the function called low_level_output() to
 * do the actuall transmission of the packet.
 *
 */
/*----------------------------------------------------------------------------*/

static err_t
sosif_output(struct netif *netif, struct pbuf *p,
	       struct ip_addr *ipaddr)
{
	struct pbuf *q;

	q = etharp_output(netif, ipaddr, p);

	if (q == NULL) {
		return ERR_MEM;
	} else {
		return low_level_output(netif, q);
	}
}

/*
 * sosif_init():
 *
 * Should be called at the beginning of the program to set up the
 * network interface.
 */

static struct netif *our_netif;

void
sosif_init(struct netif *netif)
{

	our_netif = netif;
	netif->state = NULL;

	netif->name[0] = IFNAME0;
	netif->name[1] = IFNAME1;

	netif->output = sosif_output;
	netif->linkoutput = linkoutput;
}

void
sosif_tx_freepacket (void *unused, uint32_t buffer, uint64_t key)
{
	assert(((struct pbuf*) key)->ref > 0);
	pbuf_free((struct pbuf*) key);
}


int
low_level_input(void *unused, uint32_t buffer_addr, 
		size_t size, uint64_t key)
{
	struct netif *netif = our_netif;
	struct pbuf *p = NULL, *q = NULL;
	int same;
	int j;
	struct eth_hdr *ethhdr;

	p = pbuf_alloc(PBUF_LINK, size, PBUF_EXT);
	
	if (p == NULL) {
		l4e_printf("mungiif_input: pbuf_alloc failed, "
			   "dropping packet\n");
		return -1;
	} else {
		p->headers = p->payload = (void*) (uint64_t) buffer_addr;
		
		p->len = size;

		/* set up the callback function */
		p->free = sosif_free_buffer; /* FIX */
		p->arg[1] = (void*)(uint64_t) buffer_addr;
		p->arg[2] = (void*)(uint64_t) key;
	}

#ifdef LINK_STATS
	stats.link.recv++;
#endif /* LINK_STATS */
	
	ethhdr = p->payload;
	same = 1;
	for (j = 0; j < 6; j++) {
		if (ethhdr->src.addr[j] != netif->hwaddr[j]) {
			same = 0;
			break;
		}
	}

	if (same) {
		pbuf_free(p);
		return 0;
	}

	switch (htons(ethhdr->type)) {
	case ETHTYPE_IP:
		q = etharp_ip_input(netif, p);
		pbuf_header(p, -14);
		netif->input(p, netif);
		break;
	case ETHTYPE_ARP:
		q = etharp_arp_input(netif, (struct eth_addr *)
				     &netif->hwaddr, p);
		break;
	default:
		l4e_printf("sosif: dropped packet with bogus type\n");
		pbuf_free(p);
		break;
	}
	
	if (q != NULL) {
		linkoutput(netif, q);
		pbuf_free(q);
	}

	return NULL;
}

static void
sosif_free_buffer(struct pbuf *p)
{
	tulip_freepacket(our_netif->state, (uint64_t) p->arg[1], 1536, 0);
}
