/*
  Tulip device driver
*/

#include <l4.h>
#include <l4e.h>
#include <dev.h>
#include <stddef.h>
#include <tulip.h>
#include <io.h>
#include <stdbool.h>
#include <stdlib.h>
#include <string.h>
#include "tulip_driver.h"
#include "2114n.reg.h"

/* FIXME */
#define PAGESIZE 0x1000

#define byte_swap16(x) \
         ((unsigned short)( \
                 (((unsigned short)(x) & (unsigned short)0x00ffU) << 8) | \
                 (((unsigned short)(x) & (unsigned short)0xff00U) >> 8) ))

#define byte_swap32(x) \
         ((unsigned int)( \
                 (((unsigned int)(x) & (unsigned int)0x000000ffUL) << 24) | \
                 (((unsigned int)(x) & (unsigned int)0x0000ff00UL) <<  8) | \
                 (((unsigned int)(x) & (unsigned int)0x00ff0000UL) >>  8) | \
                 (((unsigned int)(x) & (unsigned int)0xff000000UL) >> 24) ))

#define MEDIA_TP	1
#define MEDIA_AUI	2
#define MEDIA_BNC	3
#define MEDIA		MEDIA_TP /* FIXME: make this configurable/autosense */

#define D_INT		0x01
#define D_SROM_READ	0x02
#define D_EN_RCV	0x04
#define D_FEED		0x08
#define D_FREE		0x10
#define D_INIT		0x40
#define D_RXDESC        0x80

//b#define VERBOSITY	(D_FEED)
#define VERBOSITY 0

/* Seedy GNUism used in macro */
#define DEBUG(lvl, str, ...) { \
	if (VERBOSITY & lvl) l4e_printf("tulip: " str, ##__VA_ARGS__); }

static void tulip_setup_media(struct tulip *);
static void srom_read(struct tulip *);
static int setup_transmit(struct tulip *dthis, uint32_t buffer_addr, size_t size, 
			  uint64_t key, bool first, bool last, bool setup_frame);
static int tulip_receive_interrupt(struct tulip *dthis);

struct tulip *
tulip_init(bs_seq spaces, dma_handle_t dma, bus_space_t pciconf)
{
	struct tulip *dthis;
	static struct tulip tulip;
	union status status;
	union bus_mode bus_mode;
	union irq_mask irq_mask;
	union op_mode op_mode;
	int i, r;

	dthis = &tulip; /* FIXME */

	if (dthis == NULL)
		return NULL;

	dthis->dma = dma;

	dthis->pciconf = pciconf;
	assert(spaces.c >= 1);
	dthis->csr = spaces.v[0];

	dthis->rcv_state = STOPPED;

	DEBUG(D_INIT, "init called\n");

	/* reset the thing */
	bus_mode = READ_BUS_MODE();
	bus_mode.r.swr = 1;
	WRITE_BUS_MODE(bus_mode);

	dev_sleep(3); /* should be >50 PCI cycles */

	status = READ_STATUS();
#if (VERBOSITY & D_INIT)
	print_status(status);
#endif

	bus_mode = READ_BUS_MODE();
	bus_mode.r.cal = 3;
	bus_mode.r.dsl = 0;
#if defined(BIG_ENDIAN) && defined(ARCH_MIPS)
	bus_mode.r.ble = 1; /* workaround broken GT setup */
#endif
#if (VERBOSITY & D_INIT)
	print_bus_mode(bus_mode);
#endif
	WRITE_BUS_MODE(bus_mode);

	op_mode = READ_OP_MODE();
	op_mode.r.full_duplex = 1;
	op_mode.r.operating_mode = 0;
#if (VERBOSITY & D_INIT)
	print_op_mode(op_mode);
#endif
	WRITE_OP_MODE(op_mode);

	tulip_setup_media(dthis);

	srom_read(dthis); /* Read the serial rom */
	DEBUG(D_INIT, "MAC: %02x-%02x-%02x-%02x-%02x-%02x\n", 
	      dthis->rombuf[20], dthis->rombuf[21], dthis->rombuf[22],
	      dthis->rombuf[23], dthis->rombuf[24], dthis->rombuf[25]);

	dthis->mac_addr = 0;
	for (i = 0; i < 6; i++)
		dthis->mac_addr |= ((uint64_t)dthis->rombuf[20 + i]) << (i * 8);
	DEBUG(D_INIT, "MAC as 64bit is: 0x%lx\n", dthis->mac_addr);

	/* MAC address snaity checking here */

	DEBUG(D_INIT, "Setting up descriptor rings\n");

	r = get_pinnable_memory(PAGESIZE, &dthis->rcv_ring_handle);
	dthis->rcv_ring = get_addr(dthis->rcv_ring_handle);
	DEBUG(D_INIT, "rcv_ring is at %p\n", dthis->rcv_ring);
	dthis->rcv_ring_dma = setup_dma(dthis->dma, dthis->rcv_ring_handle,
					PAGESIZE, consistent);
	r = get_pinnable_memory(PAGESIZE, &dthis->send_ring_handle);
	dthis->send_ring = get_addr(dthis->send_ring_handle);
	dthis->send_ring_dma = setup_dma(dthis->dma, dthis->send_ring_handle,
					 PAGESIZE, consistent);

	DEBUG(D_INIT, "Initialising receive description ring\n");
	/* Set up the receive descriptor ring */
	for (i = 0; i < RX_NUMDESC; i++) {
		dthis->rcv_ring[i].rdes0.raw = 0;
		dthis->rcv_ring[i].rdes1.raw = 0;
		dthis->rcv_ring[i].rdes2 = 0;
		if (i == RX_NUMDESC - 1) {
			dthis->rcv_ring[i].rdes1.x.rer = 1; /* end of ring */
			dthis->rcv_ring[i].rdes3 = 0;
		} else {
			dthis->rcv_ring[i].rdes1.x.rch = 0; /* chained */
			dthis->rcv_ring[i].rdes3 = 0;
			//dthis->rcv_ring_dma +
			//	((i + 1) * sizeof(struct recv_descriptor));
		}
	}

	dthis->rcv_ring_start_pos = 0;
	dthis->rcv_ring_end_pos = 0;
	dthis->rcv_full = 1;

	/* Set up the transmit descriptor ring */
	for (i = 0; i < TX_NUMDESC; i++) {
		dthis->send_ring[i].tdes0.raw = 0;
		dthis->send_ring[i].tdes1.raw = 0;
		dthis->send_ring[i].tdes2 = 0;
		if (i == TX_NUMDESC - 1) {
			dthis->send_ring[i].tdes1.x.ter = 1; /* end of ring */
			dthis->send_ring[i].tdes3 = 0;
		} else {
			dthis->send_ring[i].tdes1.x.tch = 1; /* chained */
			dthis->send_ring[i].tdes3 = dthis->send_ring_dma +
				((i + 1) * sizeof(struct trans_descriptor));
		}
	}

	dthis->send_ring_start_pos = 0;
	dthis->send_ring_end_pos = 0;

	l4e_cache_flush();

	assert(dthis->rcv_ring_dma % 4 == 0);
	assert(dthis->send_ring_dma % 4 == 0);

	/* XXX: BROKEN - the casting to uint32_t is dodgy */
	WRITE_RCV_RING((union rcv_ring) (uint32_t) dthis->rcv_ring_dma);
	WRITE_SEND_RING((union send_ring) (uint32_t) dthis->send_ring_dma);

	DEBUG(D_INIT, "start of receive list: 0x%08x\n", dthis->rcv_ring_dma);
	DEBUG(D_INIT, "start of send    list: 0x%08x\n", dthis->send_ring_dma);
	DEBUG(D_INIT, "start of recv list VA: 0x%08x\n", dthis->rcv_ring);
	DEBUG(D_INIT, "start of send list VA: 0x%08x\n", dthis->send_ring);

	irq_mask = READ_IRQ_MASK();
	irq_mask.r.tim = 1;
	irq_mask.r.tsm = 1;
	irq_mask.r.tum = 1;
	irq_mask.r.tjm = 1;
	irq_mask.r.lpm = 1;
	irq_mask.r.unm = 1;
	irq_mask.r.rim = 1;
	irq_mask.r.rum = 1;
	irq_mask.r.rsm = 1;
	irq_mask.r.rwm = 1;
	irq_mask.r.tmm = 1;
	irq_mask.r.lfm = 1;
	irq_mask.r.sem = 1;
	irq_mask.r.erm = 1;
	irq_mask.r.aim = 1;
	irq_mask.r.nim = 1;
	WRITE_IRQ_MASK(irq_mask);

#if 0
	/* bfr_init should be some callback function */
	DEBUG(D_INIT, "bfr_init\n");

	if (bfr_init(dthis, &dthis->bufman, dthis->dma,
		     RX_NUMDESC, RX_DESCLEN, TX_NUMDESC) != 0) {
		/*free(dthis); */
		DEBUG(D_INIT, "bad\n");
		return NULL;
	}
#endif

	DEBUG(D_INIT, "done\n");

	return dthis;
}

static void
tulip_setup_media(struct tulip *dthis)
{
	union sia_txrx sia_txrx;
	union sia_conn sia_conn;
	union sia_general sia_general;

	sia_conn.data = 0;
	WRITE_SIA_CONN(sia_conn);

	sia_txrx = READ_SIA_TXRX();
	sia_txrx.r.lbk = 0; /* disable loopback */
	sia_txrx.r.mbo = 1;
	sia_txrx.r.ane = 1;
	sia_txrx.r.tas = 1;
#if (VERBOSITY & D_INIT)
	print_sia_txrx(sia_txrx);
#endif
	WRITE_SIA_TXRX(sia_txrx);

	sia_general = READ_SIA_GENERAL();
#if MEDIA == MEDIA_AUI
	sia_general.r.abm = 1; /* select AUI over BNC in AUI/BNC mode */
#else
	sia_general.r.abm = 0; /* select BNC over AUI in AUI/BNC mode */
#endif

#if (VERBOSITY & D_INIT)
	print_sia_general(sia_general);
#endif
	WRITE_SIA_GENERAL(sia_general);

	sia_conn.data = 0;
	sia_conn.r.sdm = 0xef0; /* cos the spec sez so */
#if MEDIA == MEDIA_TP
	sia_conn.r.aui = 0; /* select 10baseT mode over AUI/BNC mode */
#else
	sia_conn.r.aui = 1; /* select AUI/BNC mode over 10baseT mode */
#endif
	sia_conn.r.cac = 1; /* autoconfiguration */
	sia_conn.r.srl = 1; /* reset */
#if (VERBOSITY & D_INIT)
	print_sia_conn(sia_conn);
#endif
	WRITE_SIA_CONN(sia_conn);

	dev_sleep(10000); /* 10ms delay, as per spec */
}

/* =========================== SROM reading code ============================ */

static void delay_300ns(struct tulip *);
static void srom_idle(struct tulip *);
#define SROM_EMIT do { WRITE_SROM_MII(srom); delay_300ns(dthis); } while (0)

static void
srom_read(struct tulip *dthis)
{
	/* Stolen from NetBSD */
	unsigned idx; 
	const unsigned bitwidth = SROM_BITWIDTH;
	const unsigned cmdmask = (SROMCMD_RD << bitwidth);
	const unsigned msb = 1 << (bitwidth + 3 - 1);
	unsigned lastidx = (1 << bitwidth) - 1;

	srom_idle(dthis);

	for (idx = 0; idx <= lastidx; idx++) {
		unsigned lastbit, data, bits, bit;
		union srom_mii srom;

		srom.data = 0;
		srom.r.sr = 1; SROM_EMIT;
		srom.r.rd = 1; SROM_EMIT;
		srom.r.cs = 1; SROM_EMIT; 
		srom.r.clk = 1;
		lastbit = 0;

		for (bits = idx|cmdmask, bit = bitwidth + 3; 
		     bit > 0; bit--, bits <<= 1) {
			const unsigned thisbit = bits & msb;

			srom.r.clk = 0; SROM_EMIT;

			if (thisbit != lastbit) {
				srom.r.dout ^= 1;
				SROM_EMIT;
			} else {
				SROM_EMIT;
			}

			srom.r.clk = 1; SROM_EMIT;
			lastbit = thisbit;
		}

		srom.r.clk = 0; SROM_EMIT;

		for (data = 0, bits = 0; bits < 16; bits++) {
			union srom_mii tmp;

			data <<= 1;
			srom.r.clk = 1; SROM_EMIT;

			tmp = READ_SROM_MII();
			DEBUG(D_SROM_READ, "SROM read tmp = 0x%x\n", tmp);

			data |= tmp.r.din;
			srom.r.clk = 0; SROM_EMIT;
		}

		dthis->rombuf[idx*2] = data & 0xFF;
		dthis->rombuf[idx*2+1] = data >> 8;

		srom.data = 0;
		srom.r.sr = 1;
		srom.r.rd = 1;
		SROM_EMIT;

		srom.data = 0; SROM_EMIT;
	}

	srom_idle(dthis);
}

/* assumes 33MHz PCI (FIXME?) */
static void
delay_300ns(struct tulip *dthis)
{
	int idx;

	for (idx = (300 / 33) + 1; idx > 0; idx--)
		READ_BUS_MODE();
}

static void
srom_idle(struct tulip *dthis)
{
	unsigned int bit;
	union srom_mii srom;

	srom.data = 0;
	srom.r.sr = 1; SROM_EMIT;
	srom.r.rd = 1; SROM_EMIT;
	srom.r.cs = 1; SROM_EMIT;
	srom.r.clk = 1; SROM_EMIT;

	/*
	 * Write 25 cycles of 0 which will force the SROM to be idle.
	 */
	for (bit = 3 + SROM_BITWIDTH + 16; bit > 0; bit--) {
		srom.r.clk = 0; SROM_EMIT;
		srom.r.clk = 1; SROM_EMIT;
	}

	srom.r.clk = 0; SROM_EMIT;
	srom.r.clk = 1; SROM_EMIT;
	srom.data = 0; SROM_EMIT;
}

/* construct and send a setup frame
 * returns zero on success, nonzero on error */
static int
send_setup_frame(struct tulip *dthis)
{
	size_t size = 16 * 3 * sizeof(uint32_t);
	uint32_t *frame;
	uint32_t physaddr;
	int pos;

	frame = malloc(size);
	if (frame == NULL)
		return 0;

	memset(frame, '\0', size);

	/* MAC addresses are stored in the frame in 2-byte chunks
	 * we only care about receiving packets to our MAC and broadcast */
	/* FIXME: is this endianness-clean? */
	frame[0] = byte_swap32(dthis->mac_addr & 0xFFFF);
	frame[1] = byte_swap32((dthis->mac_addr >> 16) & 0xFFFF);
	frame[2] = byte_swap32((dthis->mac_addr >> 32) & 0xFFFF);

	frame[3] = byte_swap32(0xFFFF);
	frame[4] = byte_swap32(0xFFFF);
	frame[5] = byte_swap32(0xFFFF);

#if (VERBOSITY & D_EN_RCV)
	for (pos = 0; pos < 16; pos++)
		DEBUG(D_EN_RCV, "setup frame [%d]: 0x%x\n", pos, frame[pos]);
#endif

	/* small size doesn't work for some reason? CEG */
/*	physaddr = setup_dma(dthis->dma, new_mem_handle(frame), size, */
	physaddr = setup_dma(dthis->dma, new_mem_handle(frame), PAGESIZE,
			     mem_to_bus);
	assert(physaddr != NULL);

	//lock_lock(&dthis->send_lock);

	/* so the card can see the MAC addresses */
	l4e_cache_flush();

	pos = setup_transmit(dthis, physaddr, size, 0, false, false, true);
	if (pos == -1) {
		//teardown_dma(dthis->dma, physaddr);
		free(frame);
		return -1;
	}

	DEBUG(D_EN_RCV, "waiting for setup frame to go through...\n");

	/* poll for the frame to be `sent' */
	while (dthis->send_ring[pos].tdes0.x.own) {
		//dev_yield(); /* Yeild causes an error for some reason */
		dev_sleep(1); 
	}

	DEBUG(D_EN_RCV, "setup send done 0x%x\n",
	      dthis->send_ring[pos].tdes0.raw);

	dthis->send_ring[pos].tdes2 = NULL;
	dthis->send_ring_start_pos = pos + 1;

	//lock_unlock(&dthis->send_lock);

	//teardown_dma(dthis->dma, physaddr);
	free(frame);

	return 0;
}

int
tulip_enable_receive(struct tulip *dthis)
{
	union op_mode op_mode;
	int r;

	DEBUG(D_EN_RCV, "Enabling receive\n");

	/* since we're going to use perfect filtering, we need to construct
	 * and send a setup frame */
	r = send_setup_frame(dthis);
	assert(r == 0);
	if (r != 0)
		return r;

	op_mode = READ_OP_MODE();
//	op_mode.r.hash_perfect = 0;
//	op_mode.r.inverse_filter = 0;
	op_mode.r.start_rcv = 1;
//	op_mode.r.hash_only = 0;
	op_mode.r.pass_bad_frames = 0;
#if 0 /* receive all packets (if things are working, except filtering) */
	op_mode.r.promisc = 1;
	op_mode.r.pass_multi = 1;
#else
	op_mode.r.promisc = 0;
	op_mode.r.pass_multi = 0; /* FIXME: no multicast */
#endif
//	op_mode.r.rcv_all = 1;
//	op_mode.r.port_select = 1;

	dthis->rcv_state = RUNNING;

	WRITE_OP_MODE(op_mode);
#if (VERBOSITY & D_EN_RCV)
	print_op_mode(op_mode);
#endif

	DEBUG(D_EN_RCV, "Enabled receive\n");

	return 0;
}

/* add a frame to the transmit descriptors
 * returns descriptor used, or -1 on error */
static int
setup_transmit(struct tulip *dthis, uint32_t buffer_addr, size_t size, 
	       uint64_t key, bool first, bool last, bool setup_frame)
{
	union op_mode op_mode;
	union status status;
	int i, j;

	DEBUG(D_FEED, "Feeding packet at: %x\n", buffer_addr);
	if (size == 0) {
		UserPrint("tulip: feeding zero size packets moron!\n");
		return -1;
	}
	if (buffer_addr == 0) {
		UserPrint("tulip: feeding NULL pointer moron\n");
		return -1;
	}

	i = dthis->send_ring_end_pos;
	if (dthis->send_ring[i].tdes0.x.own) {
		UserPrint("tulip: dropping packet due to full transmit ring\n");
		return -1;
	}

	if (!setup_frame)
		dthis->transmit_keys[i] = key;

	dthis->send_ring_end_pos++;
	dthis->send_ring_end_pos %= TX_NUMDESC;

	if (first && !setup_frame)
		dthis->send_frame_start_pos = i;

	
	dthis->send_ring[i].tdes1.x.fs = !setup_frame && (first != 0);
	dthis->send_ring[i].tdes1.x.ls = !setup_frame && (last != 0);
	dthis->send_ring[i].tdes1.x.ft0 = 0;
	dthis->send_ring[i].tdes1.x.ft1 = 0;
	dthis->send_ring[i].tdes1.x.set = (setup_frame != 0);
	dthis->send_ring[i].tdes1.x.ic = !setup_frame; /* int on completion */
	dthis->send_ring[i].tdes1.x.tbs1 = size;
	dthis->send_ring[i].tdes2 = buffer_addr; /* transmit buffer address */
	dthis->send_ring[dthis->send_ring_end_pos].tdes0.raw = 0;

	if (setup_frame || last) {
		/* make the card "own" all the buffers in the frame */
		//mb();
		for (j = i; ; j = (j - 1) % TX_NUMDESC) {
			dthis->send_ring[j].tdes0.raw = 0;
			dthis->send_ring[j].tdes0.x.own = 1;
			if (setup_frame || j == dthis->send_frame_start_pos)
				break;
		}

		/* let the card see the new descriptors */
		l4e_cache_flush();

		status = READ_STATUS();
		if (setup_frame || status.r.tps == 1) {
			/* Start the transmit process */
			DEBUG(D_FEED, "Starting TX process\n");
			op_mode = READ_OP_MODE();
			op_mode.r.start_send = 1;
			WRITE_OP_MODE(op_mode);
		} else if (status.r.ts == 6 || status.r.tu == 1) {
			/* transmit suspended, kick it */
			DEBUG(D_FEED, "TX Suspend, restarting\n");
			WRITE_SEND_POLL((union send_poll)(uint32_t)0);
		}
	}

	DEBUG(D_FEED, "finished packet send\n");

	return i;
}

int
tulip_get_macaddr(struct tulip *tulip, uint8_t macaddr[6])
{
	int i;
	for (i=0; i<6; i++)
		macaddr[i] = (uint8_t) (tulip->mac_addr >> (i * 8));
	return 0;
}

static int
tulip_transmit_interrupt(struct tulip *dthis)
{
	int i;

	for (i = dthis->send_ring_start_pos; ; i = (i + 1) % TX_NUMDESC) {
		DEBUG(D_INT, "pos %d: buf %x own %d\n", i,
		      dthis->send_ring[i].tdes2,
		      dthis->send_ring[i].tdes0.x.own);

		if (i == dthis->send_ring_end_pos
		    || dthis->send_ring[i].tdes2 == NULL
		    || dthis->send_ring[i].tdes0.x.own)
			break;

		if (dthis->send_ring[i].tdes2 != NULL) {
			DEBUG(D_INT, "transmit done: freeing %d\n", i);
			
			dthis->tx_freepacket(dthis->callback_data,
					     dthis->send_ring[i].tdes2,
					     dthis->transmit_keys[i]);

			dthis->send_ring[i].tdes2 = NULL;
		}

		dthis->send_ring_start_pos = (i + 1) % TX_NUMDESC;
	}

	return 0;
}

int
tulip_interrupt_handler(struct tulip *dthis)
{
	union status status;

	DEBUG(D_INT, "interrupt\n");

	while(1) {
		status = READ_STATUS();

		if (status.r.nis == 0 &&  status.r.ais == 0) {
			break;
		}

		WRITE_STATUS(status);

		if (status.r.ti == 1) {
			DEBUG(D_INT, "A transmit interrupt\n");
			tulip_transmit_interrupt(dthis);
		}
		
		if (status.r.fbe == 1) {
			/* This should never happen */
			assert(!"tulip: Fatal bus error\n");
		}
		
		if (status.r.ru == 1 || status.r.ri == 1) {
			if (status.r.ru == 1) {
				/* Recieve buffer unavailable, so we stop */
				dthis->rcv_state = SUSPENDED;
			}
			
			if (status.r.ri == 1) {
				tulip_receive_interrupt(dthis);
			}
		}
	}
	return 0;
}

static int
tulip_receive_interrupt(struct tulip *dthis)
{
	int i, n = 0;

	for (i = dthis->rcv_ring_start_pos; ; i = (i + 1) % RX_NUMDESC) {
		DEBUG(D_RXDESC, "pos is %d, rcv_ring_end_pos is %d\n", 
		      i, dthis->rcv_ring_end_pos);

		l4e_cache_flush();

		DEBUG(D_RXDESC, "start_pos is %d, end_pos is %d\n",
		      dthis->rcv_ring_start_pos, dthis->rcv_ring_end_pos);
		
#if (VERBOSITY & D_RXDESC)
		{
			int j;
			UserPrint("r%d ", dthis->rcv_ring_start_pos);
			for (j = 0; j < RX_NUMDESC; j++)
				UserPrint("%d", dthis->rcv_ring[j].rdes0.x.own);
			UserPrint("\n");
		}
#endif

		if (dthis->rcv_ring[i].rdes0.x.own
		    || dthis->rcv_ring[i].rdes2 == NULL) {
			if (n == 0) {
				if (
				    dthis->rcv_ring[(i + 1) % RX_NUMDESC].
				    rdes0.x.own == 0) {
					assert(!"Broken network card driver");
				}
					
			} else if (dthis->rcv_ring[i].rdes0.x.own == 0) {
				UserPrint("Really evil null pointer %d\n", i);
				l4e_printf("es %d, fs %d, ls %d, "
					   "fl %d  own %d rdes2 %p %p\n", 
					   dthis->rcv_ring[i].rdes0.x.es,
					   dthis->rcv_ring[i].rdes0.x.fs,
					   dthis->rcv_ring[i].rdes0.x.ls,
					   dthis->rcv_ring[i].rdes0.x.fl,
					   dthis->rcv_ring[i].rdes0.x.own,
					   dthis->rcv_ring[i].rdes2, 
					   &dthis->rcv_ring[i].rdes2);
				assert(!"Broken network card driver\n");
			}
			break;
		}

		dthis->rcv_ring_start_pos = (i + 1) % RX_NUMDESC;

		if (!dthis->rcv_ring[i].rdes0.x.own
		    && !dthis->rcv_ring[i].rdes0.x.es
		    && dthis->rcv_ring[i].rdes0.x.fs
		    && dthis->rcv_ring[i].rdes0.x.ls) {
			uint32_t tmp = dthis->rcv_ring[i].rdes2;
			n++;
			dthis->rcv_ring[i].rdes2 = NULL;

			dthis->rx_feedpacket(dthis->callback_data,
					     tmp,
					     dthis->rcv_ring[i].rdes0.x.fl,
					     dthis->rcv_keys[i]);
		} else {
			UserPrint("tulip: error in receive: %d 0x%08x 0x%x 0x%x"
				  " 0x%x\n", i, dthis->rcv_ring[i].rdes0.raw,
				  dthis->rcv_ring[i].rdes1.x.rbs1,
				  dthis->rcv_ring[i].rdes2,
				  dthis->rcv_ring[i].rdes3);
			UserPrint("tulip: rdes1 of %d is 0x%x\n", i,
				  dthis->rcv_ring[i].rdes1.raw);
		}

	}

#if (VERBOSITY & D_RXDESC)
	UserPrint("r%d ", dthis->rcv_ring_start_pos);
	for (i = 0; i < RX_NUMDESC; i++)
		UserPrint("%d", dthis->rcv_ring[i].rdes0.x.own);
	UserPrint("\n");
	for (i = 0; i < RX_NUMDESC; i++)
		UserPrint("(%d)%p\n", i, dthis->rcv_ring[i].rdes2);
#endif

	return 0;
}


int
tulip_freepacket(struct tulip *tulip, uint32_t buffer_addr, size_t size,
		 uint32_t key)
{
	int i;
	struct tulip *dthis = tulip;

	i = tulip->rcv_ring_end_pos;

	while (tulip->rcv_ring[i].rdes0.x.own == 1) {
		i++;
		i %= RX_NUMDESC;
		if (i == tulip->rcv_ring_end_pos) {
			assert("No free descriptor slots\n");
		}
	}

	/* Got a free packet, add to our rcv list */
	DEBUG(D_FREE, "Adding recv buffer %x/%d @ %d\n", buffer_addr, size, i);

	/* required by card, should handle this better though */
	assert(size % 4 == 0 && buffer_addr % 4 == 0);
	assert(tulip->rcv_ring[i].rdes0.x.own == 0);
	/* can't handle buffers larger than this */
	if (size >= 2048)
		size = 2047;

	tulip->rcv_ring[i].rdes0.raw = 0;
	tulip->rcv_ring[i].rdes1.x.rbs1 = size;
	tulip->rcv_ring[i].rdes2 = buffer_addr;
	tulip->rcv_ring[i].rdes0.x.own = 1;

	DEBUG(D_FREE, "### recv %p %x\n", &tulip->rcv_ring[i].rdes2, 
	      tulip->rcv_ring[i].rdes2);

	tulip->rcv_ring_end_pos++;
	tulip->rcv_ring_end_pos %= RX_NUMDESC;

	if (tulip->rcv_full == 1)
		tulip->rcv_full = 0;

	if (tulip->rcv_state == SUSPENDED) {
		DEBUG(D_FREE, "Polling\n");
		WRITE_RCV_POLL((union rcv_poll)(uint32_t)0);
		tulip->rcv_state = RUNNING;
	}

	return 0;
}

int
tulip_feedpacket(struct tulip *tulip, uint32_t buffer_addr, size_t size,
	       uint8_t first, uint8_t more,
	       uint64_t key)
{
	return setup_transmit(tulip, buffer_addr, size, key, 
			      first, !more, false);
}

int 
tulip_register_callback(struct tulip *tulip,
			void *callback_data, 
			void (*tx_freepacket) (void *, uint32_t, uint64_t),
			int (*rx_feedpacket) (void *, uint32_t, 
					       size_t, uint64_t))
{
	tulip->callback_data = callback_data;
	tulip->tx_freepacket = tx_freepacket;
	tulip->rx_feedpacket = rx_feedpacket;

	return 1;
}
